/********************************************************************************************************************************/
/*																	*/
/*                                                       book		*/
/*																	*/
/*							                interactive librarian for ibm pc's	*/
/*																	*/
/*	 	    									program by Jacques Mattheij	*/
/*																	*/
/********************************************************************************************************************************/
/* FOLD START speech !												*/
/********************************************************************************************************************************/
/* work : 
   0000 23/07/91 x startup message gone
   0001 23/07/91 x filter items 8 / 9 gone
   0002 23/07/91 - 2 lines extra in text window
   0003 23/07/91 x copyright message gone
   0004 23/07/91 x & from command line
   0005 23/07/91 x help is a mess
   0006 23/07/91 x page-down key
   0007 23/07/91 x scroll errors
   0008 23/07/91 x case sensitivity on keywords off
   0009 23/07/91 x keyboard buffering
   0010 25/07/91 x screen update too slow
   0011 27/07/91 x index too small
   0012 28/07/91 x topic display
   0013 29/07/91 x cursor could be better; moving on the same line needs
                   only one line redraw
   0014 29/07/91 x screen update when cursor reaches end of page or start of
                   page is unneccesary
   0015 29/07/91 x spaces in topic and filter items need to take on background
                   colour.
   0016 31/07/91 x page up/down in topic display 
   0017 31/07/91 x page down doesn't check last logical line
   0018 31/07/91 x index display
   0019 04/08/91 x alpha keys put you in the right position in index display
   0020 04/08/91 x online help function
   0021 04/08/91 x put you at page with parameter words in title rather first
                   page containing all
   0022 04/08/91 x end user documentation in a book
   0023 06/08/91 x page up / down in main text display move too many lines
                   now break if difference between old line and new line more
                   than 16 lines
   0024 06/08/91 x page one in help keys short explanation
   0025 06/08/91 x enter key is out of order
   0026 06/08/91 x program exit after a session through a terminal doesn't
                   restore all settings properly modified qnxtc.c
   0027 25/12/93 x make indexer work with 26 separate keyword lists
                 ? input box for word to search for
                 - when the cursor moves within one line the line needs not
                   be reprinted; only the old and the new word need to be
                   reprinted
                 - home and end keys in main display
                 - separate index manager
                 - cumulative index
                 - logical or
                 - wildcards
*/

/*																	*/
/* this program is an interactive text query processor.				*/
/* it allows you to look at speccially indexed text files in a few	*/
/* different ways (basically the only difference is in the amount	*/
/* of interactive stuff you skip because you specify it in parameters	*/
/* as soon as parameters are exhausted or an error is encountered the	*/
/* program switches to interactive mode.							*/
/*																	*/
/* example 1: the maximum command line								*/
/*    get * &cd &rom &video !philips								*/
/*    which means get me from all books all articles containing key	*/
/*    words cd rom and video not containing philips					*/
/* 																	*/
/* example 2: just which books										*/
/*    get lang*														*/
/*																	*/
/* example 3: the easy way out										*/
/*    get															*/
/*																	*/
/* books should be in the directory /usr/local/lib/books or in a subdirecory	 	*/
/* specified by an environmental variable named books.				*/
/*																	*/
/* books are formatted like this :									*/
/*   pages															*/
/*     lines														*/
/*       text														*/
/*       end_of_line (cr,lf)										*/
/*     end_of_page (form_feed)										*/
/*																	*/
/* end of book is detected automatically. the end of page marker for the	*/
/* last page is optional; it is neater however to supply it.		*/
/*																	*/
/* for each book a binary index file with the same filename as the book 	*/
/* but extension .idx should be present. if not the program will squeak	*/
/* about it and prompt you about how to make one.					*/
/*																	*/
/* there's different sources of books. you can make your own using the	*/
/* index generator; or you can buy pre-written books. anything readable	*/
/* can be turned into a book as long as it adheres to the above 	*/
/* requirments.														*/
/*																	*/
/* return zaps you to the next occurence of the word under the cursor	*/
/* i      gets you an alphabetical index							*/
/* t      gets you an available title display						*/
/* b      switches to alternative books								*/
/* n      previous word in filter									*/
/* p      next     word in filter									*/
/* w      allows you to enter a wild card specifier					*/
/* update help text													*/
/*																	*/
/********************************************************************************************************************************/
/* FOLD END   speech !												*/
/* FOLD START include_files_and_defines								*/
/*
 * # include files
 */

#if 0
#define VDU_ANS          1
#endif
#define INCLUDE_SCREEN   1
#define INCLUDE_KEYBOARD 1
#include "/msc/src/lib/jmlib.h"
#include <unistd.h>

/*
#define DEBUG 1
#define DEBUG_MAIN 1
*/

/*
 * # defines
 */

#define TRUE 1													/* boolean true	*/
#define FALSE 0													/* boolean false	*/

/*
 * limits
 */

#define LINE_WIDTH     132
#define LINES_PER_PAGE 10000
#define PAGES_PER_BOOK 512										/* should be kept a multiple of 8	*/
 
#define SEL_TYPE_NOT 0											/* item shouldn't be found to include page	*/
#define SEL_TYPE_AND 1											/* item should be found to include page	*/

typedef struct {
	int type;
	unsigned char specifier[32];
	unsigned char criterium[PAGES_PER_BOOK >> 3];
} SEL_ITEM;

#define MAX_SEL_ITEMS 8

#define MAX_STACK_SIZE 6

typedef struct {
	char title [LINE_WIDTH+2];
	int n_sel_items;
	SEL_ITEM sel_items[MAX_SEL_ITEMS];
	int current_page;

	int x_offset;
	int crline;													/* current line	*/
	
	int cxpos;													/* current slot	*/
	int cypos;													/* current screen line	*/
	int pages_selected;
	unsigned char current_selection[PAGES_PER_BOOK >> 3];
} LOC_STACK_ITEM;	

LOC_STACK_ITEM loc_stack[MAX_STACK_SIZE];
int loc_stack_ptr=0;
LOC_STACK_ITEM * lsp = loc_stack;

char current_book[0x80];

char current_book_title[0x80];

typedef struct {
	char line[LINE_WIDTH+2];									/* one extra for termination 0	*/
} SCREEN_LINE;
 
/* FOLD END   include_files_and_defines								*/
/* FOLD START error messages										*/
#define ERC_BAC 0
#define ERC_BDO 1
#define ERC_COI 2
#define ERC_COO 3
#define ERC_NOM 4

char error_supplement[80] = { 0 };

char * error_msgs[] = {
	"bad argument count",
	"bad option %s",
	"cannot open %s for input",
	"cannot open %s for output",
	"no files match specifier %s"
};

/* FOLD END   error messages										*/
/* FOLD START global_storage										*/
char progname[] = "book (interactive book viewer)";
char version[]  = "2.01";

int wind_x;
int wind_y;
int wind_h;
int wind_w;

int opt_v=FALSE;

int attrib=1;													/* file attributes to search for	*/
 
int c_index  = 14;
int c_filter = 12;
int c_plain  = 7;
int c_back   = 0;
int c_error  = 12;
int c_border = 7;
int c_title  = 15;
int c_nonasc = 1;

int colour_max[] = {15,15,15,7,15,15,15,15};
char * colour_descriptors[] = {
	" tokens in index      ",
	" tokens in filter     ",
	" punctuation          ",
	" background           ",
	" error messages       ",
	" borders              ",
	" topic title          ",
	" non-ascii garbage    "
};

char spaces[] = "       ";
unsigned char blocks[] = { 219,219,219,219,219,219,219,0 };

char * colour_demos[] = {
	blocks,
	blocks,
	blocks,
	spaces,
	blocks,
	blocks,
	blocks,
	blocks
};

SCREEN_LINE (* screen)[LINES_PER_PAGE];							/* 2d representation of tree	*/

int n_lines;													/* number of lines on page	*/

int first_logical_line;
int last_logical_line;

/*
 * index file
 * filled in get_book
 */

int n_pages;													/* number of pages in this book	*/
long (* page_starts)[PAGES_PER_BOOK];							/* beginning of each page in file	*/
char * page_titles[PAGES_PER_BOOK];								/* title of each page	*/

typedef struct {
	char * p;
	int npages;
	short int * pages;
	void * next;
} TOKEN;

int n_tokens;													/* number of tokens	*/
TOKEN * tokens;													/* tokens in linked list	*/
 
char * index_m;													/* memory allocated for whole index */
char * index_p;													/* global index read pointer */


char error_message[132];
int redraw;

char token_under_cursor[0x80];

/* FOLD END   global_storage										*/
/* FOLD START subroutines											*/
trace(s)

char * s;

{
	FILE * fp;

#if defined(DEBUG) || defined(DEBUG_MAIN)
	fp = fopen("trace.txt","a+");

	fprintf(fp,"%s\n",s);

	fclose(fp);
#endif
}
talk(s)

char * s;

{
	if (opt_v) {
		printf(s);
	}
}

#if 0
#undef vdu_fcol
#undef vdu_bcol

void
vdu_fcol(int fcol)

{
}

void
vdu_bcol(int bcol)

{
}
#endif

do_help(h)

char * h;

{
	if (strcmp(current_book,"book") != 0) {
		vdu_exit();
		system(h);
		vdu_init();
		return(TRUE);
	}

	sprintf(error_message," you are already being helped ");
	return(FALSE);
}

int criterium_set(c,p)

unsigned char * c;
int p;

{
	static char bits[] = { 0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01 };

	c += p >> 3;
	if (*c & bits[p & 0x07]) {
		return(TRUE);
	}
	else {
		return(FALSE);
	}
}

add_page_to_criterium(c,p)

unsigned char * c;
int p;

{
	static char bits[] = { 0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01 };

#ifdef DEBUG
	trace("-> add_page_to_criterium");
#endif
	c += p >> 3;
	*c |= bits[p & 0x07];	
#ifdef DEBUG
	trace("<- add_page_to_criterium");
#endif
}
 
remove_page_from_criterium(c,p)

unsigned char * c;
int p;

{
	static char bits[] = { 0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01 };

#ifdef DEBUG
	trace("-> remove page from criterium");
#endif
	c += p >> 3;
	*c &= ~bits[p & 0x07];	
#ifdef DEBUG
	trace("<- remove page from criterium");
#endif
}

reset_criterium(c)

unsigned char * c;

{
	int i;

#ifdef DEBUG
	trace("-> reset criterium");
#endif
	for (i=0;i<(PAGES_PER_BOOK >> 3);i++) {
		*c++ = 0;
	}
#ifdef DEBUG
	trace("<- reset criterium");
#endif
}

full_criterium(c)

unsigned char * c;

{
	int i;

#ifdef DEBUG
	trace("-> full criterium");
#endif
	reset_criterium(c);
	for (i=0;i<n_pages;i++) {
		add_page_to_criterium(c,i);
	}
#ifdef DEBUG
	trace("<- full criterium");
#endif
}

not_criterium(c)

unsigned char * c;

{
	int i;

	for (i=0;i<(PAGES_PER_BOOK >> 3);i++) {
		*c ^= 0xff;
		c++;
	}
}
 
and_criterium(c0,c1)

unsigned char * c0;
unsigned char * c1;

{
	int i;

	for (i=0;i<(PAGES_PER_BOOK >> 3);i++) {
		*c0 &= *c1;
		c0++;
		c1++;
	}
}

 
count_pages_selected()

{
	int i;
	
	lsp->pages_selected = 0;
	for (i=0;i<n_pages;i++) {
		if (criterium_set(lsp->current_selection,i)) {
			lsp->pages_selected++;
		}
	}

	if (lsp->pages_selected == 0) {
		sprintf(error_message," empty set ");
	}
}

compute_selection()

{
	int i;

	full_criterium(lsp->current_selection);

	for (i=0;i<lsp->n_sel_items;i++) {
		and_criterium(lsp->current_selection,lsp->sel_items[i].criterium);
	}

	count_pages_selected();
}

add_pages_to_criterium(c,token)

unsigned char * c;
TOKEN * token;

{
	short int * ip;	
	int np;

	ip = token->pages;
	np = token->npages;

	while (np--) {
		add_page_to_criterium(c,*ip++);
	}
}
 
void 
str_tolower(c)

char * c;

{
	while (*c) {
		if (isalpha(*c)) {
			*c = *c | 0x20;
		}
		c++;
	}
}

int add_sel_item(type,specifier)

int type;
char * specifier;

{
	int found;
	int n;
	int l;
	TOKEN * p;
	int np;

	str_tolower(specifier);

	if (lsp->n_sel_items == MAX_SEL_ITEMS) {
		sprintf(error_message," no room for another selection item ");
		return(FALSE);
	}

	lsp->sel_items[lsp->n_sel_items].type = type;	

	reset_criterium(lsp->sel_items[lsp->n_sel_items].criterium);
	strcpy(lsp->sel_items[lsp->n_sel_items].specifier,specifier);
	
	found = FALSE;

	p = tokens;
	n = n_tokens;
	while (n-- && (!found)) {
		if (strcmp(specifier,p->p) == 0) {
			found = TRUE;
			add_pages_to_criterium(lsp->sel_items[lsp->n_sel_items].criterium,p);
		}

		p = p->next;
	}

	if (type == SEL_TYPE_NOT) {
		not_criterium(lsp->sel_items[lsp->n_sel_items].criterium);
	}

	if (found) {
		lsp->n_sel_items++;

		compute_selection();

		return(TRUE);
	}
	compute_selection();
	sprintf(error_message," specifier %s not in index ",specifier);
	return(FALSE);
}

char * get_token_at_address(sp)

char * sp;

{
	static char buffer[0x80];

	char * dp;

	dp = buffer;

	while (is_token_char(*sp)) {
		if (isalpha(*sp)) {
			*dp++ = tolower(*sp);
			sp++;
		}
		else {
			*dp++ = *sp++;
		}
	}
	*dp++ = 0;
	return(buffer);
}


int in_sel_items(p)

char * p;

{
	char * t;
	int n;

	t = get_token_at_address(p);
	n = lsp->n_sel_items;
	while (n) {
		n--;

		if (strcmp(lsp->sel_items[n].specifier,t) == 0) {
			return(TRUE);
		}
	}
	return(FALSE);
}

delete_sel_item(i)

int i;

{
	if (i>=lsp->n_sel_items) {
		sprintf(error_message," filter element %d is empty ",i);
		return;
	}

	while (i < lsp->n_sel_items-1) {
		lsp->sel_items[i] = lsp->sel_items[i+1];
		i++;
	}
	lsp->n_sel_items--;
	compute_selection();
}

interpret_command(c)

char * c;

{
	int type;

	switch (*c) {
		case '!':
			type = SEL_TYPE_NOT;
			c++;
			break;
		default:
			type = SEL_TYPE_AND;
			break;
	}

	if (!add_sel_item(type,c)) {
		return(FALSE);
	}

	return(TRUE);
}

int get_line(fp,line)

FILE * fp;
int line;

{
	char * p;

#ifdef DEBUG
	trace("-> get line");
#endif

	p = (char *) &((*screen)[line]);

	if (fgets(p,LINE_WIDTH,fp)== 0) {
		if (first_logical_line == -1) {
			first_logical_line = line;
		}
		last_logical_line = line - 1;

		*p = 0;
#ifdef DEBUG
		trace("   eof");
		trace("-> get line");
#endif
		return(FALSE);
	}

	while (*p != 13 && *p != '\n') {	
		if (is_token_char(*p)) {
			if (first_logical_line == -1) {
					first_logical_line = line;
			}
			last_logical_line = line;
		}

		if (*p == 0x0c) {
			*p = 0;
			return(FALSE);
		}
		p++;
	}
	*p = 0;
#ifdef DEBUG
	trace("   non eof");
	trace("<- get line");
#endif
	return(TRUE);
}
 
int get_title(fp)

FILE * fp;

{
	char * p;

	p = lsp->title;

	fgets(p,LINE_WIDTH,fp);

	while (*p != 13 && *p != '\n' && *p != 0) {	
/*		fprintf(stderr,"%c",*p);
*/		p++;
	}
	*p = 0;
}

build_screen(page)

int page;

{
	char name_and_ext[0x80];

	FILE * fp;

#ifdef DEBUG
	trace("-> build screen page");
#endif
	sprintf(name_and_ext,"/usr/local/lib/books/%s.bk",current_book);

	fp = fopen(name_and_ext,"rb");

	if (fp == NULL) {
		sprintf(error_message," could not open book %s ",name_and_ext);
#ifdef DEBUG
		trace("-> error no book file");
#endif
		return(FALSE);
	}

	first_logical_line = -1;
	last_logical_line  = -1;

	n_lines = 0;

	fseek(fp,(*page_starts)[page],0);

	get_title(fp);

	while (get_line(fp,n_lines)) {
		n_lines++;
		if (n_lines == LINES_PER_PAGE) {
			sprintf(error_message," page too long ");
			break;
		}
	}

	n_lines = last_logical_line + 1;

	fclose(fp);

#ifdef DEBUG
	trace("<- build screen");
#endif
}
 
int is_token_char(c)

int c;

{
	return(isascii(c) && (isalpha(c) || isdigit(c)));
}

show_screen_line(y,line,forcursor)

int y;
int line;
int forcursor;

{
	int in_cursor;
	unsigned char * p;
	int x;
	int c;
	int skipped;

	if (!forcursor) {
		vdu_tab(0,y);				

		vdu_eeol();												/* start with a clean line	*/
	}

	vdu_tab(0,y);												/* back to the beginning	*/

	in_cursor = FALSE;

	if (line < n_lines) {
		p = (char *) &((*screen)[line]);

		x = 0;

		while (x < lsp->x_offset) {
			if (!*p) {	
				break;			
			}
			p++;
			x++;
		}

		skipped = 0;
		while (*p == ' ') {
			skipped++;			
			p++;
			x++;
		}

		if (skipped) {
			vdu_tab(skipped,y);
		}

		vdu_fcol(c_index);

		while (*p) {
			if (*p == ' ') {		
				vdu_fcol(c_index);
				vdu_raw(*p++);
				x++;
				continue;
			}
			if (is_token_char(*p)) {
				if (in_sel_items(p)) {
					if (vdu_monochrome) {
					}
					else {
						vdu_fcol(c_filter);
					}
				}
				else {
					if (vdu_monochrome) {
/*						vdu_highlight(); */
					}
					else {
						vdu_fcol(c_index);
					}
				}
				if (y == lsp->cypos && x == lsp->cxpos) {
					in_cursor = TRUE;
					vdu_reverse();
				}
				while (is_token_char(*p)) {
					vdu_raw(*p++);
					x++;
				}
				if (in_cursor) {
					vdu_normal();
				}
				continue;
			}
			if (isascii(*p)) {
				vdu_fcol(c_plain);
			}
			else {
				vdu_fcol(c_nonasc);
			}
#if defined(M_I386) && !defined(GNUDOS)
			if (*p >= 0xb3 && *p <= 0xda) {
				/* ibm grahphics characters */
				switch (*p++) {
					case 0xb3:
					case 0xba:
						vdu_raw('|');
						break;
					case 0xc4:
					case 0xcd:
						vdu_raw('-');
						break;
					default:
						vdu_raw('+');
						break;
				}
			}
			else {
				vdu_raw(*p++);
			}
#else
			vdu_raw(*p++);
#endif
			x++;
		}
	}
}
 
updscr()

{
	int line;
	int y;

	vdu_twin(wind_x,wind_y,wind_w,wind_h);
	if (lsp->cypos > lsp->crline) {
		lsp->cypos = lsp->crline;
	}
	line = lsp->crline - lsp->cypos;

	for (y=0;y<wind_h;y++) {
		if (line < n_lines) {
			show_screen_line(y,line++,FALSE);
		}
		else {
			vdu_tab(0,y);
			vdu_eeol();
		}
	}
}

get_token_under_cursor()

{
	char * dp;
	char * sp;
	
	sp = &(*screen)[lsp->crline];

	sp += lsp->cxpos;

	dp = token_under_cursor;

	while (is_token_char(*sp)) {
		if (isalpha(*sp)) {
			*dp++ = tolower(*sp);
			sp++;
		}
		else {
			*dp++ = *sp++;
		}
	}
	*dp++ = 0;
}
 
int under_cursor_is_token_char()

{
	char * p;

	p = &(*screen)[lsp->crline];

	return(is_token_char(*(p + lsp->cxpos)));
}

int cursor_at_end_of_line()

{
	char * p;

	p = &(*screen)[lsp->crline];

	if (lsp->cxpos >= strlen(p)) {
		return(TRUE);
	}

	p+= lsp->cxpos;
	while (*p) {
		if (is_token_char(*p++)) {
			return(FALSE);
		}
	}
	return(TRUE);
}

/*
 * put cursor at first token character to be encountered
 * starting from current_cursor position.
 * if no token characters are encountered the routine will end at
 * with the cursor at the beginning of the text
 */

left_until_token_or_sot()

{
	char * p;

	p = &((*screen)[lsp->crline]);

	while (!is_token_char(*(p + lsp->cxpos))) {
		if (lsp->cxpos == 0) {
			if (lsp->crline == 0) {
				return;
			}

			lsp->cypos--;
			lsp->crline--;			

			p = &((*screen)[lsp->crline]);

			lsp->cxpos = strlen(p);
 
		}
		else {
			lsp->cxpos--;
		}
	}
}
 
/*
 * move the cursor left while it is on token characters.
 * leaves the cursor on the first character of the token
 */

left_while_token()

{
	char * p;

	p = &((*screen)[lsp->crline]);

	while (lsp->cxpos && is_token_char(*(p + lsp->cxpos))) {
		lsp->cxpos--;
	}
	if (lsp->cxpos == 0) {
		if (!is_token_char(*p)) {
			lsp->cxpos++;
		}
	}
	else {
		lsp->cxpos++;
	}
}

to_next_token()

{
	char * p;

	p = &((*screen)[lsp->crline]);
 
	while (is_token_char(*(p + lsp->cxpos))) {
		if (*(p+lsp->cxpos+1) == 0) {
			if (lsp->crline < last_logical_line) {
				lsp->crline++;
				lsp->cypos++;
				p = &((*screen)[lsp->crline]);
				lsp->cxpos = 0;
				break;
			}
			left_while_token();
			return;
		}
		lsp->cxpos++;
	}
 
	while (!is_token_char(*(p+lsp->cxpos))) {
		if (*(p+lsp->cxpos+1) == 0) {
			if (lsp->crline < n_lines - 1) {
				lsp->crline++;
				lsp->cypos++;
				p = &((*screen)[lsp->crline]);
				lsp->cxpos = 0;
			}
			else {
				left_until_token_or_sot();
				left_while_token();
				return;
			}
		}
		else {
			lsp->cxpos++;
		}
	}

}

to_prev_token()

{
	if (lsp->cxpos == 0) {
		if (lsp->crline == 0) {
			return;
		}

		lsp->cypos--;
		lsp->crline--;			

		lsp->cxpos = strlen(&((*screen)[lsp->crline]));
	}
	else {
		lsp->cxpos--;
	}
	left_until_token_or_sot();
	left_while_token();
}

up()

{
	int x;
	int line;

	if (lsp->crline > first_logical_line) {
		x = lsp->cxpos;
		line = lsp->crline;

		while (lsp->crline > first_logical_line) {
			left();
			get_token_under_cursor();
			if (x >= lsp->cxpos && x <= lsp->cxpos + strlen(token_under_cursor)) {
				return;
			}
		}
	}
	lsp->cypos = first_logical_line;
	redraw = TRUE;
	left();
}

left()

{
	to_prev_token();

	if (!under_cursor_is_token_char()) {
		right();
	}

	if (lsp->cxpos < lsp->x_offset) {		
		lsp->x_offset = lsp->cxpos - 10;
		if (lsp->x_offset < 0) {
			lsp->x_offset = 0;
		}
		redraw = TRUE;
 	}
}

/*
 * returns true if failed
 */

int right()

{
	to_next_token();
	if (lsp->cxpos >= (lsp->x_offset + wind_w)) {
		lsp->x_offset = (lsp->cxpos + 10) - wind_w;
		redraw = TRUE;
	}
}

down()

{
	if (lsp->crline != last_logical_line) {
		lsp->crline++;
		lsp->cypos++;

		while (cursor_at_end_of_line()) {
			lsp->cypos++;
			lsp->crline++;
			if (lsp->crline == last_logical_line) {
				break;
			}
		}

		if (cursor_at_end_of_line()) {
			lsp->cxpos = strlen(&(*screen)[lsp->crline])-1;
		}

		if (under_cursor_is_token_char()) {
			left_while_token();
		}
		else {
			to_next_token();
		}

		if (lsp->cypos >= wind_h) {
			redraw = TRUE;
		}
	}
	else {
		right();
	}
}

sright()

{
	if (lsp->pages_selected == 0) {
		return;
	}
	do {
		lsp->current_page++;
		if (lsp->current_page == n_pages) {
			lsp->current_page = 0;
		}
	} while (criterium_set(lsp->current_selection,lsp->current_page)==0);
	build_screen(lsp->current_page);
	output_title();
	redraw = TRUE;
	lsp->crline = 0;
	lsp->x_offset = 0;
	lsp->cxpos = 0;
	lsp->cypos = 0;
	left();
}


sleft()

{
	if (lsp->pages_selected == 0) {
		return;
	}
	do {
		lsp->current_page--;
		if (lsp->current_page == -1) {
			lsp->current_page = n_pages-1;
		}
	} while (criterium_set(lsp->current_selection,lsp->current_page)==0);
	build_screen(lsp->current_page);
	output_title();
	redraw = TRUE;
	lsp->crline = 0;
	lsp->x_offset = 0;
	lsp->cxpos = 0;
	lsp->cypos = 0;
	left();
}

int
strnicmp(s1, s2, n)		/* <0 for <, 0 for ==, >0 for > */

char *s1;
char *s2;
size_t n;

{
	register char *scan1;
	register char *scan2;
	register size_t count;

	scan1 = s1;
	scan2 = s2;
	count = n + 1;		/* extra so predecrement tests remainder */
	while (--count > 0 && *scan1 != '\0' && toupper(*scan1) == toupper(*scan2)) {
		scan1++;
		scan2++;
	}
	if (count <= 0) {
		return(0);
	}
	/* The following case analysis is necessary so that characters which
	 * look negative collate low against normal characters but high
	 * against the end-of-string NUL. */
	if (*scan1 == '\0' && *scan2 == '\0')
		return(0);
	else if (*scan1 == '\0')
		return(-1);
	else if (*scan2 == '\0')
		return(1);
	else
		return(toupper(*scan1) - toupper(*scan2));
}


int 
instr(char * p0,char * p1) 

{
	int i;
	int l0,l1;
	char * p2;
	char * p3;

	l0 = strlen(p0);
	l1 = strlen(p1);

	i = 0;
	while (l0 >= l1) {
		if (strnicmp(p0,p1,l1)==0) {
			return(i);
		}

		p0++;
		l0--;
		i++;
	}
	return(-1);
}

int
basic_enter()

{
	int started;

	started = lsp->current_page;
 
	do {
		lsp->current_page++;
		if (lsp->current_page == n_pages) {
			lsp->current_page = 0;
		}

		if (instr(page_titles[lsp->current_page],token_under_cursor) > -1) {
			build_screen(lsp->current_page);
			output_title();
			redraw = TRUE;
			lsp->crline = 0;
			lsp->x_offset = 0;
			lsp->cxpos = 0;
			lsp->cypos = 0;
			left();

			return (TRUE);
		}
	} while (started != lsp->current_page);
	return (FALSE);
}
void
enter()

{
	push_location();

	reset_filter();

	get_token_under_cursor();
 
	if (!basic_enter()) {
		pull_location();	
	}
}

highlight()

{
	if (vdu_monochrome) {
		vdu_highlight();
	}
	else {
		vdu_fcol(c_index);
	}
}

highlight_off()

{
	if (vdu_monochrome) {
		vdu_highlight_off();
	}
	else {
		vdu_fcol(c_plain);
	}
}

newl()

{
	vdu_cr();
	vdu_lf();
}

first_highlighted(s,x,y)

char * s;
int x;
int y;

{
	vdu_tab(x,y);
	if (vdu_monochrome) {
		vdu_highlight();
	}
	else {
		vdu_fcol(c_index);
	}
	vdu_raw(*s++);
	if (vdu_monochrome) {
		vdu_highlight_off();
	}
	else {
		vdu_fcol(c_plain);
	}
	vdu_str(s);
}

init_screen()

{
	int i;
	int * ip;
	char s[0x80];

	vdu_dwin();

	if (vdu_monochrome) {
		ip = &c_index;
		for (i=0;i<8;i++) {
			if (*ip > 1) {
				*ip = 1;
			}
			else {
				*ip = 0;
			}
		}
		if (c_border == c_back) {
			c_border != c_back;
		}
	}

	vdu_bcol(c_back);
	vdu_fcol(c_border);

	vdu_cuof();
 
/*	vdu_clsc();	*/												/* clear screen	*/
	vdu_normal();
	vdu_highlight_off();
	win_dwi(0,0,vdu_screen_x,3);
	win_dwi(0,vdu_screen_y-4,vdu_screen_x,4);	
	vdu_reverse();
	vdu_tab(3,0);

	vdu_tab(3,vdu_screen_y-1);

	sprintf(s," Book %s (C) MCS 1992 ",version);

	vdu_str(s);

	vdu_tab(3,0);
	vdu_raw(' ');
	vdu_str(current_book_title);
	vdu_raw(' ');

	vdu_normal();
	win_dhl(2,1,vdu_screen_x - 2);
}

/* FOLD END   subroutines											*/
/* FOLD START index display											*/
int index_cypos;
int index_crline;

TOKEN * 
get_token_no(t)

int t;

{
	TOKEN * tp;

	tp = tokens;

	while (t--) {
		tp = tp->next;
	}

	return(tp);
}
 
show_index_screen_line(y,t)

int y;
int t;

{
	TOKEN * tp;

	tp = tokens;

	vdu_tab(0,y);												

	vdu_fcol(c_index);
	vdu_bcol(c_back);

	if (t == index_crline) {
		vdu_reverse();
	}
	while (t--) {
		tp = tp->next;
	}

	vdu_str(tp->p);
	vdu_eeol();
	vdu_normal();
}

do_index_updscr()

{
	int y;
	int line;

	if (index_cypos > index_crline) {
		index_cypos = index_crline;
	}
	line = index_crline - index_cypos;

	for (y=0;y<wind_h;y++) {
		if (line < n_tokens) {
			show_index_screen_line(y,line++);
		}
		else {
			vdu_tab(0,y);
			vdu_eeol();
		}
	}
}

do_index_title()

{
	vdu_twin(1,1,vdu_screen_x-2,1);

	vdu_bcol(c_back);
	vdu_fcol(c_title);

	vdu_tab(0,0);

	vdu_eeol();

	win_cnt(" Index Display",0,0,vdu_screen_x);

	vdu_twin(wind_x,wind_y,wind_w,wind_h);
 
}

do_index()

{
	static int first = TRUE;
	int old_cypos;
	int old_crline;
	int k;
	TOKEN * tp;
	int l;
 
	do_index_title();

	do_index_updscr();
	do {
		output_error();
		vdu_twin(wind_x,wind_y,wind_w,wind_h);
 
		vdu_tab(vdu_screen_x-1,vdu_screen_y-2);
		if (first) {
			k = 'a';
			first = FALSE;
		}
		else {
			vdu_refresh();
			key_flush();
			k = key_rdch();
			if (k == 13) {
				k = '\n';
			}
		}
		old_cypos = index_cypos;
		old_crline = index_crline;
		if (isalpha(k)) {
			k = tolower(k);
			tp = tokens;
			l = 0;
			while (l < n_tokens) {
				if (*tp->p == k) {
					index_cypos = 0;
					index_crline = l;
					do_index_updscr();
					break;					
				}
				tp = tp->next;
				l++;
			}
			if (l == n_tokens) {
				sprintf(error_message,"no item with that letter");
			}
		}
		else {
			switch (k) {
				case KEY_F1:
					if (do_help("book book index display")) {

						redraw_screen();
						do_index_title();

						vdu_twin(wind_x,wind_y,wind_w,wind_h);
	
						do_index_updscr();
					}

					break;
 
				case '!':
					tp = get_token_no(index_crline);
					add_sel_item(SEL_TYPE_NOT,tp->p);
					sright();
					output_filter_items();
					old_cypos = -1;
					redraw = TRUE;
					break;
				case '\n':
				case '&':
					k = '&';
					tp = get_token_no(index_crline);
					add_sel_item(SEL_TYPE_AND,tp->p);
					output_filter_items();
					sleft();
					sright();
					old_cypos = -1;
					redraw = TRUE; 
					break;
				case KEY_DOWN:
					if (index_crline < (n_tokens-1)) {
						index_crline++;
						index_cypos++;
						if (index_cypos >= wind_h) {
							index_cypos = wind_h-1;
							do_index_updscr();
						}
						else {
							show_index_screen_line(old_cypos,old_crline);
							show_index_screen_line(index_cypos,index_crline);
						}
					}
					break;
				case KEY_UP:
					if (index_crline > 0) {
						index_crline--;
						index_cypos--;
						if (index_cypos < 0) {
							index_cypos = 0;
							do_index_updscr();
						}
						else {
							show_index_screen_line(old_cypos,old_crline);
							show_index_screen_line(index_cypos,index_crline);
						}
					}
					break;
				case KEY_SUP:
					if (index_crline > 0) {
						index_crline -= 16;
						index_cypos -= 16;
						if (index_crline < 0) {
							index_crline = 0;
						}
						if (index_cypos < 0) {
							index_cypos = 0;
							do_index_updscr();
						}
						else {
							show_index_screen_line(old_cypos,old_crline);
							show_index_screen_line(index_cypos,index_crline);
						}
					}
					break;
				case KEY_SDOWN:
					if (index_crline < (n_tokens - 1)) {
						index_crline += 16;
						index_cypos  += 16;
						if (index_crline > n_tokens-1) {
							index_crline = n_tokens-1;
						}
						if (index_cypos > n_tokens-1) {
							index_cypos = n_tokens-1;
						}
						if (index_cypos >= wind_h) {
							index_cypos = wind_h-1;
							do_index_updscr();
						}
						else {
							show_index_screen_line(old_cypos,old_crline);
							show_index_screen_line(index_cypos,index_crline);
						}
					}
					break;
			}
		}
	} while (k != 0x1b && k != '&' && k != '!');

	redraw = TRUE;
	output_title();
/*	redraw_screen();
*/
}

/* FOLD END   topics display										*/

/* FOLD START topics display										*/
int topics_cypos;
int topics_crline;

show_topics_screen_line(y,line)

int y;
int line;

{
	vdu_tab(0,y);												

	vdu_fcol(c_title);
	vdu_bcol(c_back);

	if (line == topics_crline) {
		vdu_reverse();
	}
	vdu_str(page_titles[line]);
	vdu_eeol();
	vdu_normal();
}

do_topics_updscr()

{
	int y;
	int line;

	if (topics_cypos > topics_crline) {
		topics_cypos = topics_crline;
	}
	line = topics_crline - topics_cypos;

	for (y=0;y<wind_h;y++) {
		if (line < n_pages) {
			show_topics_screen_line(y,line++);
		}
		else {
			vdu_tab(0,y);
			vdu_eeol();
		}
	}
}

do_topics_title()

{
	vdu_twin(1,1,vdu_screen_x-2,1);

	vdu_bcol(c_back);
	vdu_fcol(c_title);

	vdu_tab(0,0);

	vdu_eeol();

	win_cnt(" Topics Display",0,0,vdu_screen_x);

	vdu_twin(wind_x,wind_y,wind_w,wind_h);
}

do_topics()

{
	int old_cypos;
	int old_crline;
	int k;
 
	do_topics_title();
 
	do_topics_updscr();
	do {
		vdu_tab(vdu_screen_x-1,vdu_screen_y-2);
		vdu_refresh();
		key_flush();
		k = key_rdch();
		if (k == 13) {
			k = '\n';
		}
		old_cypos = topics_cypos;
		old_crline = topics_crline;
		switch (k) {
				case KEY_F1:
					if (do_help("book book topic display")) {

						redraw_screen();
						do_topics_title();

						vdu_twin(wind_x,wind_y,wind_w,wind_h);

						do_topics_updscr();
					}
					break;



			case KEY_DOWN:
				if (topics_crline < (n_pages-1)) {
					topics_crline++;
					topics_cypos++;
					if (topics_cypos >= wind_h) {
						topics_cypos = wind_h-1;
						do_topics_updscr();
					}
					else {
						show_topics_screen_line(old_cypos,old_crline);
						show_topics_screen_line(topics_cypos,topics_crline);
					}
				}
				break;
			case KEY_UP:
				if (topics_crline > 0) {
					topics_crline--;
					topics_cypos--;
					if (topics_cypos < 0) {
						topics_cypos = 0;
						do_topics_updscr();
					}
					else {
						show_topics_screen_line(old_cypos,old_crline);
						show_topics_screen_line(topics_cypos,topics_crline);
					}
				}
				break;
			case KEY_SUP:
				if (topics_crline > 0) {
					topics_crline -= 16;
					topics_cypos -= 16;
					if (topics_crline < 0) {
						topics_crline = 0;
					}
					if (topics_cypos < 0) {
						topics_cypos = 0;
						do_topics_updscr();
					}
					else {
						show_topics_screen_line(old_cypos,old_crline);
						show_topics_screen_line(topics_cypos,topics_crline);
					}
				}
				break;
			case KEY_SDOWN:
				if (topics_crline < (n_pages - 1)) {
					topics_crline += 16;
					topics_cypos  += 16;
					if (topics_crline > n_pages-1) {
						topics_crline = n_pages-1;
					}
					if (topics_cypos > n_pages-1) {
						topics_cypos = n_pages-1;
					}
					if (topics_cypos >= wind_h) {
						topics_cypos = wind_h-1;
						do_topics_updscr();
					}
					else {
						show_topics_screen_line(old_cypos,old_crline);
						show_topics_screen_line(topics_cypos,topics_crline);
					}
				}
				break;
		}
	} while (k != '\n' && k != 0x1b);

	if (k == '\n') {
		lsp->current_page = topics_crline;
		build_screen(topics_crline);
		lsp->crline = 0;
		lsp->x_offset = 0;
		lsp->cxpos = 0;
		lsp->cypos = 0;
		left();
	}
	else {
		pull_location();
	}

	redraw = TRUE;
	output_title();
/*	redraw_screen();
*/
}

/* FOLD END   topics display										*/
/* FOLD START colour configuration display							*/	
do_configure()

{
	int cypos = 0;

	int backup[8];
	int key;
	int i;
	int * ip;

/*	vdu_screen_push();
*/
	ip = &c_index;
	for (i=0;i<8;i++) {
		backup[i] = *ip++;
	}
 
	vdu_dwin();

	vdu_clsc();													/* clear screen	*/

	vdu_fcol(c_border);
	vdu_bcol(c_back);

	vdu_normal();

	win_dwi(0,0,vdu_screen_x,vdu_screen_y);	
	vdu_reverse_on = TRUE;
	win_cnt(" book (c) MCS 1992 ",0,0,vdu_screen_x);
	vdu_reverse_on  = FALSE;

	win_cnt(" Configuration Screen ",0,3,vdu_screen_x);

	highlight();
	win_cnt("up / down select; left / right modify; esc / return exit",0,vdu_screen_y - 3,vdu_screen_x);
	highlight_off();

	vdu_twin(4,5,vdu_screen_x - 8,vdu_screen_y - 12);

	vdu_clsc();

	do {
		for (i=0;i<8;i++) {
			vdu_fcol(c_title);
			vdu_bcol(c_back);
			vdu_tab(3,3+i);

			if (i == cypos) {
				vdu_reverse();
			}
			vdu_str(colour_descriptors[i]);
			if (i == cypos) {
				vdu_normal();
			}
			vdu_raw(' ');
			vdu_fcol(backup[i]);
			vdu_bcol(backup[i]);
			vdu_str(colour_demos[i]);
		}

		vdu_tab(vdu_screen_x-1,vdu_screen_y-2);
		vdu_refresh();

		key_flush();
		key = key_rdch();
		if (key == 13) {
			key = '\n';
		}

		switch(key) {
			case KEY_UP:
			case KEY_SUP:
				cypos--;
				if (cypos < 0) {
					cypos = 7;
				}
				break;
			case KEY_DOWN:
			case KEY_SDOWN:
				cypos++;
				if (cypos > 7) {
					cypos = 0;
				}
				break;
			case KEY_LEFT:
#ifndef M_I386
			case KEY_SLEFT:
#endif
				backup[cypos]--;
				if (backup[cypos] < 0) {
					backup[cypos] = colour_max[cypos];
				}
				break;
			case KEY_RIGHT:
#ifndef M_I386
			case KEY_SRIGHT:
#endif
				backup[cypos]++;
				if (backup[cypos] > colour_max[cypos]) {
					backup[cypos] = 0;
				}
				break;
		}
	} while (key != '\n' && key != 0x1b);

	if (key == '\n') {
		ip = &c_index;
		for (i=0;i<8;i++) {
			*ip++ = backup[i];
		}
		init_screen();
		updscr();
		output_title();
		sprintf(error_message," ");
		output_error();
		output_error();
		output_filter_items();
		output_pages_selected();

		put_book_colours();
	}
	else {
		vdu_bcol(c_back);
		vdu_fcol(c_plain);
	}
	redraw_screen();
}

/* FOLD END   colour configuration display							*/
/* FOLD START main_display											*/
output_title()

{
	int i;
	char s[132];

#ifdef DEBUG
	trace("-> output_title");
#endif

	if (vdu_monochrome) {
		vdu_highlight();
	}
	else {
		vdu_fcol(c_title);
	}

	vdu_dwin();

	vdu_tab(1,1);

	for (i=1;i<vdu_screen_x-1;i++) {
		vdu_raw(' ');
	}
 
	lsp->title[vdu_screen_x - 10] = 0;

	sprintf(s,"%s",lsp->title);

	if (strlen(s) >= vdu_screen_x) {
		s[vdu_screen_x] = 0;
	}
	win_cnt(s,0,1,vdu_screen_x);

	if (vdu_monochrome) {
		vdu_highlight_off();
	}
	else {
		vdu_fcol(c_plain);
	}
#ifdef DEBUG
	trace("<- output_title");
#endif
}
 
output_pages_selected()

{
	char s[80];

	vdu_dwin();
	
	vdu_tab(55,vdu_screen_y - 1);	

	vdu_fcol(c_border);
	vdu_reverse();
 
/*	vdu_str(" Pages : ");
*/
	sprintf(s," %3d Pages of %3d left ",lsp->pages_selected,n_pages);

	vdu_str(s);

	vdu_normal();
	win_dhl(vdu_cursor_y,vdu_cursor_x,vdu_screen_x-2-vdu_cursor_x);
	vdu_twin(wind_x,wind_y,wind_w,wind_h);
	
}
 
output_stack_items()

{
	char s[10];

	vdu_dwin();
	
	vdu_tab(32,vdu_screen_y - 1);	

	vdu_fcol(c_border);
	vdu_reverse();

	vdu_str(" Bookmarks : ");

	sprintf(s,"%d ",loc_stack_ptr);

	vdu_str(s);

	vdu_normal();

	win_dhl(vdu_cursor_y,vdu_cursor_x,56 - vdu_cursor_x);
	vdu_twin(wind_x,wind_y,wind_w,wind_h);
}

output_filter_items()

{
	int l;
	int x;
	int y;
	char s[10];
	int i,j;

	l = (vdu_screen_x - 2) / 4;
	l--;

	vdu_dwin();
	
	vdu_fcol(c_filter);

	y = vdu_screen_y - 3;
	x = 2;

	for (i=0;i<MAX_SEL_ITEMS;i++) {
		if (x == 2) {
			vdu_tab(x-1,y);
		}
		else {
			vdu_tab(x,y);
		}
		for (j=0;j<l+1;j++) {
			vdu_raw(' ');
		}
 
		vdu_tab(x,y);

		sprintf(s,"(%1d) ",i);
		vdu_str(s);

		if (i<lsp->n_sel_items) {			
			switch (lsp->sel_items[i].type) {
				case SEL_TYPE_NOT:
					vdu_raw('!');
					break;
				case SEL_TYPE_AND:
					vdu_raw('&');
					break;
			}
			vdu_str(lsp->sel_items[i].specifier);
		}
		x += l;
		if (x + l >= vdu_screen_x) {
			x = 2;
			y++;
		}
	}
	vdu_twin(wind_x,wind_y,wind_w,wind_h);
}

output_error()

{
	static int escape_visible=FALSE;

	vdu_dwin();

	vdu_bcol(c_back);
	vdu_fcol(c_border);

	if (error_message[0]) {
		escape_visible = FALSE;
		win_dhl(vdu_screen_y - 4,1,vdu_screen_x - 2);
		if (vdu_monochrome) {
			vdu_highlight();
		}
		else {
			vdu_fcol(c_error);
		}
		win_cnt(error_message,0,vdu_screen_y - 4,vdu_screen_x);
		error_message[0] = 0;
		if (vdu_monochrome) {
			vdu_highlight_off();
		}
		else {
			vdu_fcol(c_plain);
		}
	}
	else {
		if (!escape_visible) {

			escape_visible = TRUE;
			win_dhl(vdu_screen_y - 4,1,vdu_screen_x - 2);
			if (strcmp(current_book,"book") == 0) {
				win_cnt(" Press ESC to quit ",0,vdu_screen_y - 4,vdu_screen_x);
			}
			else {
				win_cnt(" Press ESC to quit - F1 for help ",0,vdu_screen_y - 4,vdu_screen_x);
			}
		}
	}
}

redraw_screen()

{
	init_screen();
	output_stack_items();
	output_title();
	output_pages_selected();
	output_filter_items();
	vdu_dwin();
	error_message[0] = 32;
	error_message[1] = 0;
	output_error();
	output_error();
	redraw = TRUE;
}

int
push_location()

{
	if (loc_stack_ptr < MAX_STACK_SIZE -1) {
		lsp++;
		loc_stack_ptr++;
		loc_stack[loc_stack_ptr] = loc_stack[loc_stack_ptr-1];
		output_stack_items();
		output_pages_selected();
		return(TRUE);
	}
	sprintf(error_message,"no more bookmarks available");
	return(FALSE);
}

pull_location()

{
	if (loc_stack_ptr) {
		lsp--;
		loc_stack_ptr--;
		redraw = TRUE;
		build_screen(lsp->current_page);
		redraw_screen();
	}
}
 
reset_filter()

{
	lsp->n_sel_items = 0;
	compute_selection();
	output_filter_items();										/* reset the filter	*/
}

interact()

{
	int old_pages_selected=-1;
	int x;
	int old_x_offset;
	int old_crline;
	int old_cypos;
	int key=0;

#ifdef DEBUG_MAIN
	trace("in interact");
#endif

	key_init();

	init_screen();

	lsp->current_page = n_pages-1;
	sright();

	output_filter_items();
	output_stack_items();

	if (lsp->sel_items[0].specifier[0]) {
		strcpy(token_under_cursor,lsp->sel_items[0].specifier);
		basic_enter();
	}

	updscr();

	while (key != 27) {
		vdu_dwin();
/*		sprintf(error_message,"last logical line = %d crline = %d",last_logical_line,lsp->crline);
*/
/*		sprintf(error_message,"current line %d",lsp->crline);
*/
		output_error();
		if (lsp->pages_selected != old_pages_selected) {
			output_pages_selected();
			old_pages_selected = lsp->pages_selected;
		}
	
		vdu_tab(vdu_screen_x-1,vdu_screen_y-2);
		vdu_refresh();
		key_flush();
		key = key_rdch();

		if (key == 13) {
			key = '\n';
		}

		redraw = FALSE;
		old_x_offset = lsp->x_offset;
		old_crline = lsp->crline;
		old_cypos  = lsp->cypos;
		vdu_twin(wind_x,wind_y,wind_w,wind_h);
		switch(key) {
			case 'I':
			case 'i':
				do_index();
				output_filter_items();
				old_cypos = -1;
				redraw = TRUE;
				break;
			case 'T':
			case 't':
				if (push_location()) {
					reset_filter();
					vdu_twin(wind_x,wind_y,wind_w,wind_h);
					do_topics();
					vdu_twin(wind_x,wind_y,wind_w,wind_h);
					old_cypos = -1;
				}
				break;
			case KEY_F1:
				if (do_help("book book")) {
					old_cypos = -1;
					redraw = TRUE;
					redraw_screen();
				}
				break;
			case 'C':
			case 'c':
				do_configure();
				vdu_twin(wind_x,wind_y,wind_w,wind_h);
				old_cypos = -1;
				break;
			case KEY_UP:
				up();
				break;
			case KEY_DOWN:
				down();
				break;
			case KEY_LEFT:
				left();
				break;
			case KEY_RIGHT:
				right();
				break;
#if defined(M_I386) && !defined(GNUDOS)
			case KEY_SDOWN:
				old_cypos = -1;
				sright();
				break;
			case KEY_SUP:
				old_cypos = -1;
				sleft();
				break;
#else
			case KEY_SRIGHT:
				old_cypos = -1;
				sright();
				break;
			case KEY_SLEFT:
				old_cypos = -1;
				sleft();
				break;
			case KEY_SUP:
				for (x=0;x<16;x++) {
					up();
					if ((old_crline - lsp->crline) >= 16) {
						break;
					}
				}
				break;
			case KEY_SDOWN:
				for (x=0;x<16;x++) {
					down();
					if ((lsp->crline - old_crline) >= 16) {
						break;
					}
				}
				break;
#endif
			case KEY_CUP:
/*				home(); 
				redraw = TRUE;
				old_cypos = -1;  */
				break;
			case KEY_CDOWN:
/*				home(); */
				redraw = TRUE;
				break;
			case 'R'-64:
				reset_filter();
				old_cypos = -1;
				redraw = TRUE;
				break;

			case '!':
				get_token_under_cursor();
				add_sel_item(SEL_TYPE_NOT,token_under_cursor);
				sright();
				output_filter_items();
				old_cypos = -1;
				redraw = TRUE;
				break;
			case '&':
				get_token_under_cursor();
				add_sel_item(SEL_TYPE_AND,token_under_cursor);
				output_filter_items();
				old_cypos = -1;
				redraw = TRUE;
				break;
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
				delete_sel_item(key - 48);
				output_filter_items();
				old_cypos = -1;
				redraw = TRUE;
				break;
			case 0x20:
				push_location();

				break;
			case 27:
				if (loc_stack_ptr) {
					pull_location();
					key = 0;
					old_cypos = -1;
				}
				break;
			case '\n':
				enter();
				old_cypos = -1;
				break;
			default:
				key=-1;
		}

		if (key != -1) {
			if (redraw) {
				if (lsp->cypos > wind_h) {
					lsp->cypos = wind_h - 1;
				}
				else {
					/*
					 * put a scroll here
					 */
				}
			}
			if (lsp->cypos >= wind_h) {
				lsp->cypos = wind_h - 1;
				redraw = TRUE;
			}
			if (lsp->cypos < 0) {
				lsp->cypos = 0;
				redraw = TRUE;
			}
			if (redraw) {
				if ((old_cypos != lsp->cypos) || (old_crline != lsp->crline)) {
					updscr();
				}
			}
			else {
				if ((old_cypos != lsp->cypos) || (old_crline != lsp->crline)) {
					show_screen_line(old_cypos,old_crline,TRUE);
				}
				show_screen_line(lsp->cypos,lsp->crline,TRUE);
			}
		}
	}

	key_exit();

#ifdef DEBUG_MAIN
	trace("uit interact");
#endif

}

/* FOLD END   main_display											*/
/* FOLD START book_initialization									*/
get_string(fp,p)

FILE * fp;
char * p;

{
	do {
		fread(p,1,1,fp);
		if (feof(fp)) {
			break;
		}
		if (*p == '\n') {
			break;
		}
		if (*p != 13) {
			p++;
		}
	} while (TRUE);

/*	fread(p,1,1,fp);*/	/* skip lf */
	*p = 0;
#ifdef DEBUG
	trace("end of get string");
#endif
}

get_book_title(fp)

FILE * fp;

{
	get_string(fp,current_book_title);
}

long book_col_pos;

get_book_colours(fp)

FILE * fp;

{
	fread(&c_index ,1,2,fp);
	fread(&c_filter,1,2,fp);
	fread(&c_plain ,1,2,fp);
	fread(&c_back  ,1,2,fp);
	fread(&c_error ,1,2,fp);
	fread(&c_border,1,2,fp);
	fread(&c_title ,1,2,fp);
	fread(&c_nonasc,1,2,fp);
}

put_book_colours()

{
	int bw;
	FILE * fp;

	char name_and_ext[0x80];

	sprintf(name_and_ext,"/usr/local/lib/books/%s.idx",current_book);

	fp = fopen(name_and_ext,"r+b");

	if (fp == NULL) {
		sprintf(error_message," can't open index file %s ",name_and_ext);
		return(FALSE);
	}

	fseek(fp,0l,0);
	
	fwrite(&c_index ,1,2,fp);
	fwrite(&c_filter,1,2,fp);
	fwrite(&c_plain ,1,2,fp);
	fwrite(&c_back  ,1,2,fp);
	fwrite(&c_error ,1,2,fp);
	fwrite(&c_border,1,2,fp);
	fwrite(&c_title ,1,2,fp);
	fwrite(&c_nonasc,1,2,fp);

	fclose(fp);
}

get_topic_title(FILE * fp,int page)

{
	char s[0x100];

	fseek(fp,(*page_starts)[page],0);

	get_string(fp,s);	

	page_titles[page] = malloc(strlen(s)+1);

	strcpy(page_titles[page],s);

}

get_token(tp)

TOKEN * tp;

{
	int l;

#if 1
	l = *((int *)index_p);
	index_p += 4;
	tp->p = index_p;
	index_p += l + 1;
	tp->npages = *((int *)index_p);
	index_p += 4;
	tp->pages = (short int *)index_p;
	index_p += sizeof(short int) * tp->npages;
#else
	fread(&l,1,sizeof(int),fp);									/* token length	*/		
	tp->p = malloc(l+1);
	memset(tp->p,0,l+1);
	fread(tp->p,1,l+1,fp);										/* get token	*/	
	fread(&tp->npages,1,sizeof(int),fp);
	tp->pages = malloc(tp->npages * sizeof(short int));
	fread(tp->pages,tp->npages,sizeof(short int),fp);
#endif
}

/*
 * load a books index and check if the book itself is there
 */

int get_book(name)

char * name;

{
	int page;
	char * read_pointer;
	int pages;
	int i;
	long size;
	FILE * book_fp;
	FILE * ind_fp;
	TOKEN * last_token;
 
	char name_and_ext[0x80];
#ifdef DEBUG
	trace("in get book");
#endif
	sprintf(name_and_ext,"/usr/local/lib/books/%s.bk",name);
	book_fp = fopen(name_and_ext,"rb");

	if (book_fp == NULL) {
		sprintf(error_message," can't open book file %s ",name_and_ext);
		return(FALSE);
	}
 
	sprintf(name_and_ext,"/usr/local/lib/books/%s.idx",name);

	ind_fp = fopen(name_and_ext,"rb");

	if (ind_fp == NULL) {
		sprintf(error_message," can't open index file %s ",name_and_ext);
		return(FALSE);
	}
#ifdef DEBUG_MAIN
	trace("getting colours");
#endif
	get_book_colours(ind_fp);
#ifdef DEBUG_MAIN
	trace("getting title");
#endif
	get_book_title(ind_fp);

#ifdef DEBUG_MAIN
	trace("getting number of pages");
#endif
 
	fread(&n_pages,1,sizeof(int),ind_fp);						/* number of pages	*/

#ifdef DEBUG_MAIN
	{
		char s[0x80];
		sprintf(s,"number of pages %d",n_pages);
		trace(s);
	}
	trace("getting page starts");
#endif
	page_starts = malloc(sizeof(long) * n_pages);
	fread(page_starts,n_pages,sizeof(long),ind_fp);				/* read page starts	*/

#ifdef DEBUG_MAIN
	trace("getting number of tokens");
#endif
	fread(&n_tokens,1,sizeof(int),ind_fp);						/* number of tokens	*/

	fprintf(stderr,"index wordt geladen\n");

#ifdef DEBUG_MAIN
	trace("getting tokens");
#endif
	tokens = malloc(sizeof(TOKEN));
	last_token = tokens;

	{	
		off_t size;
		off_t pos;

		pos = ftell(ind_fp);
		fseek(ind_fp,0L,SEEK_END);
		size = ftell(ind_fp) - pos;
		fseek(ind_fp,pos,SEEK_SET);
		index_m = malloc(size);
		fread(index_m,1,size,ind_fp);
		index_p = index_m;
	}

	for (i=0;i<n_tokens;i++) {	
		if ((i % 1000) == 0) {
			fprintf(stderr,".");
		}
		get_token(last_token);		
		last_token->next = malloc(sizeof(TOKEN));
		last_token = last_token->next;
	}

#ifdef DEBUG_MAIN
	trace("full criterium");
#endif
	full_criterium(lsp->current_selection);

	lsp->n_sel_items = 0;
	lsp->current_page = 0;

	strcpy(current_book,name);

#ifdef DEBUG_MAIN
	trace("getting topic titles");
#endif

	fprintf(stderr,"\ntitels worden geladen\n");

	for (page = 0;page<n_pages;page++) {
		if ((page % 20) == 0) {
			fprintf(stderr,".");
		}
		get_topic_title(book_fp,page);
	}

	fclose(ind_fp);
	fclose(book_fp);
#ifdef DEBUG_MAIN
	trace("got book");
#endif

	vdu_clsc();

	return(TRUE);
}

/* FOLD END   book_initialization									*/
/* FOLD START book_exit												*/
forget_book()

{
	int page;

	for (page=0;page<n_pages;page++) {
		free(page_titles[page]);
	}

/*	_dos_freemem(FP_SEG(index_ptr));  */
	free(index_m);
			/* release memory occupied by index file	*/
}

/* FOLD END   book_exit												*/	
/* FOLD START book_select											*/
/*
 * interactively select a book of the available ones
 */

select_book()

{
}

/* FOLD END   book_select											*/




 

 
 
 
 
 
 
/* FOLD START main line												*/
main(argc, argv)

int argc;														/* number of arguments	*/
char *argv[];													/* pointer to argument pointers	*/

{
	int not_found;
	int e;

	char * s;
 
/*	printf(progname);
	printf(" version ");
	printf(version);
	printf(" (c) j.a. Mattheij 1988 / 1989 / 1990 / 1991\n");
*/
	vdu_init();
 
	vdu_nwrp();

	vdu_clsc();

	vdu_highlight_off();

#if 0
	vdu_screen_x = 131;
#endif

#ifdef DEBUG
	system("rm trace.txt");
#endif

	lsp->n_sel_items=0;
	lsp->current_page=0;

	lsp->x_offset=0;
	lsp->crline=0;												/* current line	*/
	
	lsp->cxpos=0;												/* current slot	*/
	lsp->cypos=0;												/* current screen line	*/
	lsp->pages_selected=0;

	screen = malloc(LINES_PER_PAGE * sizeof(SCREEN_LINE));
 
	wind_x = 0;
	wind_y = 3;
	wind_h = vdu_screen_y - 7;
	wind_w = vdu_screen_x ;

	/*
	 * parse options
	 */
	argv++;
	while (--argc > 0 && ((*argv)[0] == '-' || (*argv)[0] == '/')) {
		for (s = argv[0]+1; *s != '\0'; s++) {
			switch (toupper(*s)) {
				case 'V':
					opt_v^=TRUE;
					break;
				default:
/*					sprintf(error_supplement,"%c",*s);
					ERR_LONGJMP(ERL_SYN,ERC_BDO); */
					break;
			}
		}
		argv++;
	}
 
	switch(argc) {
		case 0:
			if (select_book()) {
				compute_selection();
				interact();
				forget_book();
			}
			break;
		default:
			if (get_book(*argv++)) {
				argc--;
				while (argc) {
					if (interpret_command(*argv++)) {
						argc--;
					}
					else {
						break;
					}
				}
				compute_selection();
				interact();
			}
			else {
				if (select_book()) {
					compute_selection();
					interact();
					forget_book();
				}	
			}
			break;
	}

	free(screen);

	vdu_exit();
}
/* FOLD END   main line												*/


 
